import json
from datetime import timedelta

from ajax_helpers.utils import random_string
from django.conf import settings
from django.forms import CharField, ModelChoiceField, NumberInput, IntegerField
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.views.generic import TemplateView
from django_datatables.columns import DateColumn, MenuColumn, DateTimeColumn
from django_datatables.datatables import DatatableExcludedRow
from django_datatables.helpers import DUMMY_ID, row_link
from django_datatables.widgets import DataTableReorderWidget
from django_menus.menu import HtmlMenu, MenuItem
from django_modals.fields import FieldEx
from django_modals.form_helpers import HorizontalNoEnterHelper
from django_modals.helper import modal_button, modal_button_method
from django_modals.modals import Modal, ModelFormModal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE
from django_modals.widgets.select2 import Select2, Select2Multiple

from advanced_report_builder.columns import ReportBuilderNumberColumn
from advanced_report_builder.data_merge.utils import DataMergeUtils
from advanced_report_builder.data_merge.widget import DataMergeWidget
from advanced_report_builder.filter_query import FilterQueryMixin
from advanced_report_builder.models import (
    CalendarReport,
    CalendarReportDataSet,
    CalendarReportDescription,
    ReportType,
)
from advanced_report_builder.utils import crispy_modal_link_args, get_report_builder_class
from advanced_report_builder.views.charts_base import ChartJSTable
from advanced_report_builder.views.datatables.utils import DescriptionColumn
from advanced_report_builder.views.helpers import QueryBuilderModelForm
from advanced_report_builder.views.modals_base import QueryBuilderModalBase
from advanced_report_builder.views.report import ReportBase


class CalendarTable(ChartJSTable):
    DATE_COLUMNS = (DateColumn, DateTimeColumn)

    def get_table_array(self, request, results):
        result_processes = self.get_result_processes()
        for p in result_processes:
            p.setup_results(request, self.page_results)
        for c in self.columns:
            c.setup_results(request, self.page_results)
        results_list = []
        for data_dict in results:
            try:
                for p in result_processes:
                    p.row_result(data_dict, self.page_results)
                row_data = []
                for column in self.columns:
                    if isinstance(column, DescriptionColumn):
                        row_data.append(column.row_result(data_dict, self.page_results, columns=self.columns))
                    elif isinstance(column, self.DATE_COLUMNS) and hasattr(column, 'iso_row_result'):
                        row_data.append(column.iso_row_result(data_dict, self.page_results))
                    else:
                        row_data.append(column.row_result(data_dict, self.page_results))

                results_list.append(row_data)
            except DatatableExcludedRow:
                pass
        return results_list


class CalendarView(DataMergeUtils, ReportBase, FilterQueryMixin, TemplateView):
    number_field = ReportBuilderNumberColumn
    template_name = 'advanced_report_builder/calendar/report.html'
    chart_js_table = CalendarTable

    def __init__(self, *args, **kwargs):
        self.chart_report = None
        super().__init__(*args, **kwargs)

    def setup_menu(self):
        super().setup_menu()
        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()
        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(report=self.report, dashboard_report=self.dashboard_report),
        )

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.chart_report = self.report.calendarreport
        self.dashboard_report = kwargs.get('dashboard_report')
        self.enable_edit = kwargs.get('enable_edit')
        return super().dispatch(request, *args, **kwargs)

    def view_filter_extra(self, query, table):
        if table.extra_query_filter:
            query = query.filter(table.extra_query_filter)
        return self.view_filter(query, table)

    def get_calendar_events(self, base_model, calendar_report_data_set, lanes, label=None, extra_query_filter=None):
        table = self.chart_js_table(model=base_model)

        report_builder_class = get_report_builder_class(
            model=base_model, report_type=calendar_report_data_set.report_type
        )
        table_indexes = ['start_date_field', 'end_date_field']
        table.add_columns(calendar_report_data_set.start_date_field)
        if calendar_report_data_set.end_date_type == CalendarReportDataSet.END_DATE_TYPE_FIELD:
            table.add_columns(calendar_report_data_set.end_date_field)
        elif calendar_report_data_set.end_date_type == CalendarReportDataSet.END_DATE_TYPE_DURATION_FIELD:
            _, _, _, start_field_col_field = self.get_field_details(
                base_model=base_model,
                field=calendar_report_data_set.start_date_field,
                report_builder_class=report_builder_class,
            )
            _, _, _, end_duration_col_field = self.get_field_details(
                base_model=base_model,
                field=calendar_report_data_set.end_duration_field,
                report_builder_class=report_builder_class,
            )

            class DurationEndDateColumn(DateColumn):
                def col_setup(self):
                    self.field = [start_field_col_field, end_duration_col_field]

                def row_result(self, data, _page_result):
                    start_date = data.get(self.model_path + start_field_col_field)
                    end_duration = data.get(self.model_path + end_duration_col_field)
                    if end_duration is None or end_duration <= 0:
                        end_duration = 3600
                    end_date = start_date + timedelta(seconds=end_duration)
                    try:
                        date = end_date.strftime('%d/%m/%Y')
                        time_str = end_date.strftime('%H:%M')
                        return date + ' ' + time_str
                    except AttributeError:
                        return ''

            table.add_columns(DurationEndDateColumn(column_name='EndDate'))

        elif calendar_report_data_set.end_date_type == CalendarReportDataSet.END_DATE_TYPE_DURATION_FIXED:
            _, _, _, start_field_col_field = self.get_field_details(
                base_model=base_model,
                field=calendar_report_data_set.start_date_field,
                report_builder_class=report_builder_class,
            )

            class FixedDurationEndDateColumn(DateColumn):
                def row_result(self, data, _page_result):

                    start_date = data.get(self.model_path + start_field_col_field)
                    end_duration = calendar_report_data_set.end_duration
                    if end_duration is None:
                        end_duration = 3600

                    end_date = start_date + timedelta(seconds=end_duration)
                    try:
                        date = end_date.strftime('%d/%m/%Y')
                        time_str = end_date.strftime('%H:%M')
                        return date + ' ' + time_str
                    except AttributeError:
                        return ''

            table.add_columns(FixedDurationEndDateColumn(column_name='EndDate'))

        if (
            calendar_report_data_set.display_type
            in (
                CalendarReportDataSet.DISPLAY_TYPE_NAME_AND_DESCRIPTION,
                CalendarReportDataSet.DISPLAY_TYPE_HEADING_ONLY,
            )
            and calendar_report_data_set.heading_field is not None
        ):
            table_indexes.append('heading')
            table.add_columns(calendar_report_data_set.heading_field)

        if calendar_report_data_set.background_colour_field is not None:
            table_indexes.append('background_colour')
            table.add_columns(calendar_report_data_set.background_colour_field)

        if (
            calendar_report_data_set.display_type
            in (
                CalendarReportDataSet.DISPLAY_TYPE_NAME_AND_DESCRIPTION,
                CalendarReportDataSet.DISPLAY_TYPE_DESCRIPTION_ONLY,
            )
            and calendar_report_data_set.calendar_report_description is not None
        ):
            description = calendar_report_data_set.calendar_report_description.description
            columns, column_map = self.get_data_merge_columns(
                base_model=base_model, report_builder_class=report_builder_class, html=description, table=table
            )
            description_column_name = (
                'description'
                if calendar_report_data_set.display_type == CalendarReportDataSet.DISPLAY_TYPE_NAME_AND_DESCRIPTION
                else 'heading'
            )
            table_indexes.append(description_column_name)
            table.add_columns(
                DescriptionColumn(
                    column_name=description_column_name, field='', html=description, column_map=column_map
                )
            )
            table.add_columns(*columns)

        table.table_options['indexes'] = table_indexes

        if calendar_report_data_set.link_field and self.kwargs.get('enable_links'):
            _, col_type_override, _, _ = self.get_field_details(
                base_model=base_model,
                field=calendar_report_data_set.link_field,
                report_builder_class=report_builder_class,
            )
            field = col_type_override.field[0] if isinstance(col_type_override.field, list) else 'id'
            if field not in table.columns:
                table.add_columns(f'.{field}')
            table.has_link = True
            table.table_options['row_href'] = row_link(col_type_override.url, field)
        else:
            table.has_link = False

        table.datatable_template = 'advanced_report_builder/calendar/middle.html'

        table.query_data = calendar_report_data_set.query_data
        table.extra_query_filter = extra_query_filter
        table.view_filter = self.view_filter_extra
        lanes.append({'datatable': table, 'label': label, 'calendar_report_data_set': calendar_report_data_set})

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['title'] = self.get_title()
        calendar_report_data_sets = self.chart_report.calendarreportdataset_set.all()
        lanes = []
        headings = []

        for calendar_report_data_set in calendar_report_data_sets:
            base_model = calendar_report_data_set.get_base_model()
            self.get_calendar_events(
                base_model=base_model, calendar_report_data_set=calendar_report_data_set, lanes=lanes
            )
        view_type = None
        if self.dashboard_report and self.dashboard_report.options is not None:
            view_type = self.dashboard_report.options.get('calendar_view_type')
            if view_type is not None:
                view_type = int(view_type)

        context['view_type'] = self.chart_report.get_view_type_for_calendar(view_type=view_type)
        context['calendar_report'] = self.chart_report
        context['headings'] = headings
        context['lanes'] = lanes
        context['field_id'] = random_string()
        return context

    def view_filter(self, query, table):
        if not table.query_data:
            return query

        return self.process_query_filters(query=query, search_filter_data=table.query_data)

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    def pod_report_menu(self):
        return [
            MenuItem(
                f'advanced_report_builder:calendar_modal,pk-{self.chart_report.id}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            ),
            *self.duplicate_menu(request=self.request, report_id=self.chart_report.id),
        ]

    def pod_dashboard_edit_menu(self):
        return [
            MenuItem(
                f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            )
        ]

    # noinspection PyMethodMayBeStatic
    def queries_menu(self, report, dashboard_report):
        return []


class CalendarModal(ModelFormModal):
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = CalendarReport
    widgets = {'report_tags': Select2Multiple}
    ajax_commands = ['datatable', 'button']

    form_fields = ['name', 'notes', 'report_tags', 'height', 'view_type']

    def form_setup(self, form, *_args, **_kwargs):
        org_id = self.object.id if hasattr(self, 'object') else None
        form.fields['notes'].widget.attrs['rows'] = 3

        if org_id is not None:
            lane_menu_items = [
                MenuItem(
                    f'advanced_report_builder:calendar_data_set_duplicate_modal,pk-{DUMMY_ID}',
                    menu_display='Duplicate',
                    css_classes='btn btn-sm btn-outline-dark',
                    font_awesome='fas fa-clone',
                ),
                MenuItem(
                    f'advanced_report_builder:calendar_data_set_modal,pk-{DUMMY_ID}',
                    menu_display='Edit',
                    css_classes='btn btn-sm btn-outline-dark',
                    font_awesome='fas fa-pencil',
                ),
            ]

            form.fields['data_sets'] = CharField(
                required=False,
                label='Data Sets',
                widget=DataTableReorderWidget(
                    model=CalendarReportDataSet,
                    order_field='order',
                    fields=[
                        '_.index',
                        '.id',
                        'name',
                        MenuColumn(
                            column_name='menu',
                            field='id',
                            column_defs={'orderable': False, 'className': 'dt-right'},
                            menu=HtmlMenu(self.request, 'button_group').add_items(*lane_menu_items),
                        ),
                    ],
                    attrs={'filter': {'calendar_report__id': self.object.id}},
                ),
            )

            description_menu_items = [
                MenuItem(
                    f'advanced_report_builder:calendar_description_duplicate_modal,pk-{DUMMY_ID}',
                    menu_display='Duplicate',
                    css_classes='btn btn-sm btn-outline-dark',
                    font_awesome='fas fa-clone',
                ),
                MenuItem(
                    f'advanced_report_builder:calendar_description_modal,pk-{DUMMY_ID}',
                    menu_display='Edit',
                    css_classes='btn btn-sm btn-outline-dark',
                    font_awesome='fas fa-pencil',
                ),
            ]

            form.fields['descriptions'] = CharField(
                required=False,
                label='Descriptions',
                widget=DataTableReorderWidget(
                    model=CalendarReportDescription,
                    order_field='order',
                    fields=[
                        '_.index',
                        '.id',
                        'name',
                        MenuColumn(
                            column_name='menu',
                            field='id',
                            column_defs={'orderable': False, 'className': 'dt-right'},
                            menu=HtmlMenu(self.request, 'button_group').add_items(*description_menu_items),
                        ),
                    ],
                    attrs={'filter': {'calendar_report__id': self.object.id}},
                ),
            )

            modal_button('Custom Close', 'close', 'btn-warning')

            return [
                *self.form_fields,
                crispy_modal_link_args(
                    'advanced_report_builder:calendar_data_set_modal',
                    'Add Data Set',
                    'calendar_report_id-',
                    self.object.id,
                    div=True,
                    div_classes='form-buttons',
                    button_classes='btn btn-primary',
                    font_awesome='fa fa-plus',
                ),
                'data_sets',
                crispy_modal_link_args(
                    'advanced_report_builder:calendar_description_modal',
                    'Add Description',
                    'calendar_report_id-',
                    self.object.id,
                    div=True,
                    div_classes='form-buttons',
                    button_classes='btn btn-primary',
                    font_awesome='fa fa-plus',
                ),
                'descriptions',
            ]

    def datatable_sort(self, **kwargs):
        form = self.get_form()
        widget = form.fields[kwargs['table_id'][3:]].widget
        _model = widget.attrs['table_model']
        current_sort = dict(_model.objects.filter(calendar_report=self.object.id).values_list('id', 'order'))
        for s in kwargs['sort']:
            if current_sort[s[1]] != s[0]:
                o = _model.objects.get(id=s[1])
                o.order = s[0]
                o.save()
        return self.command_response('')



    def post_save(self, created, form):
        if created:
            self.modal_redirect(self.request.resolver_match.view_name, slug=f'pk-{self.object.id}-new-True')
        else:
            url_name = getattr(settings, 'REPORT_BUILDER_DETAIL_URL_NAME', '')
            if url_name and self.slug.get('new'):
                url = reverse(url_name, kwargs={'slug': self.object.slug})
                self.command_response('redirect', url=url)


class CalendarDataSetForm(QueryBuilderModelForm):
    cancel_class = 'btn-secondary modal-cancel'

    class DurationWidget(NumberInput):
        crispy_kwargs = {'appended_text': 'Seconds', 'input_size': 'input-group-sm'}

    def clean(self):
        cleaned_data = super().clean()
        start_date_field = cleaned_data['start_date_field']
        if start_date_field is None:
            self.add_error('start_date_field', 'Please select a start period')
        end_date_type = cleaned_data['end_date_type']
        if (end_date_type == CalendarReportDataSet.END_DATE_TYPE_FIELD and
                cleaned_data['end_date_field'] is None):
            self.add_error('end_date_field', 'Please select a end period')
        elif (end_date_type == CalendarReportDataSet.END_DATE_TYPE_DURATION_FIELD and
              cleaned_data['end_duration_field'] is None):
            self.add_error('end_duration_field', 'Please select a end period')
        elif (end_date_type == CalendarReportDataSet.END_DATE_TYPE_DURATION_FIXED and
              (cleaned_data['end_duration'] is None or cleaned_data['end_duration'] == '')):
            self.add_error('end_duration', 'Please select a end period')
        return cleaned_data

    class Meta:
        model = CalendarReportDataSet
        fields = [
            'name',
            'report_type',
            'display_type',
            'heading_field',
            'calendar_report_description',
            'background_colour_field',
            'link_field',
            'start_date_field',
            'end_date_type',
            'end_date_field',
            'end_duration_field',
            'end_duration',
            'query_data',
        ]
        widgets = {
            'display_type': Select2,
            'end_date_type': Select2,
        }

    end_duration = IntegerField(widget=DurationWidget, required=False)


class CalendarDataSetModal(QueryBuilderModalBase):
    template_name = 'advanced_report_builder/calendar/modal.html'
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = CalendarReportDataSet
    helper_class = HorizontalNoEnterHelper
    form_class = CalendarDataSetForm

    def form_setup(self, form, *_args, **_kwargs):
        if 'data' in _kwargs:
            heading_field = _kwargs['data'].get('heading_field')
            start_date_field = _kwargs['data'].get('start_date_field')
            end_date_field = _kwargs['data'].get('end_date_field')
            end_duration_field = _kwargs['data'].get('end_duration_field')
            report_type_id = _kwargs['data'].get('report_type')
            background_colour_field = _kwargs['data'].get('background_colour_field')
            link_field = _kwargs['data'].get('link_field')
            calendar_report_description = CalendarReportDescription.objects.filter(
                pk=_kwargs['data'].get('calendar_report_description')
            ).first()
            report_type = get_object_or_404(ReportType, id=report_type_id)

        else:
            heading_field = form.instance.heading_field
            report_type = form.instance.report_type
            start_date_field = form.instance.start_date_field
            end_date_field = form.instance.end_date_field
            end_duration_field = form.instance.end_duration_field
            background_colour_field = form.instance.background_colour_field
            link_field = form.instance.link_field
            calendar_report_description = form.instance.calendar_report_description

        form.fields['calendar_report_description'].widget = Select2(attrs={'ajax': True})
        if calendar_report_description is not None:
            form.fields['calendar_report_description'].widget.select_data = [
                {'id': calendar_report_description.id, 'text': calendar_report_description.name}
            ]
        form.fields['calendar_report_description'].label = 'Description'
        form.fields['calendar_report_description'].required = False

        form.fields['report_type'] = ModelChoiceField(queryset=ReportType.objects.all(), widget=Select2, required=False)

        form.add_trigger(
            'display_type',
            'onchange',
            [
                {
                    'selector': '#div_id_calendar_report_description',
                    'values': {CalendarReportDataSet.DISPLAY_TYPE_HEADING_ONLY: 'hide'},
                    'default': 'show',
                },
                {
                    'selector': '#div_id_heading_field',
                    'values': {CalendarReportDataSet.DISPLAY_TYPE_DESCRIPTION_ONLY: 'hide'},
                    'default': 'show',
                },
            ],
        )
        form.add_trigger(
            'end_date_type',
            'onchange',
            [
                {
                    'selector': '#div_id_end_date_field',
                    'values': {CalendarReportDataSet.END_DATE_TYPE_FIELD: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': '#div_id_end_duration_field',
                    'values': {CalendarReportDataSet.END_DATE_TYPE_DURATION_FIELD: 'show'},
                    'default': 'hide',
                },
                {
                    'selector': '#div_id_end_duration',
                    'values': {CalendarReportDataSet.END_DATE_TYPE_DURATION_FIXED: 'show'},
                    'default': 'hide',
                },
            ],
        )
        self.setup_field(
            field_type='all',
            form=form,
            field_name='heading_field',
            selected_field_id=heading_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='date',
            form=form,
            field_name='start_date_field',
            selected_field_id=start_date_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='date',
            form=form,
            field_name='end_date_field',
            selected_field_id=end_date_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='number',
            form=form,
            field_name='end_duration_field',
            selected_field_id=end_duration_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='colour',
            form=form,
            field_name='background_colour_field',
            selected_field_id=background_colour_field,
            report_type=report_type,
        )

        self.setup_field(
            field_type='link', form=form, field_name='link_field', selected_field_id=link_field, report_type=report_type
        )

        return (
            'name',
            'report_type',
            'display_type',
            'heading_field',
            'calendar_report_description',
            'start_date_field',
            'end_date_type',
            'end_date_field',
            'end_duration_field',
            'end_duration',
            'background_colour_field',
            'link_field',
            FieldEx('query_data', template='advanced_report_builder/query_builder.html'),
        )

    def select2_heading_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='all', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_order_by_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='django_order', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_start_date_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='date', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_end_date_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='date', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_end_duration_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='number', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_link_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='link', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_background_colour_field(self, **kwargs):
        return self.get_fields_for_select2(
            field_type='colour', report_type=kwargs['report_type'], search_string=kwargs.get('search')
        )

    def select2_calendar_report_description(self, **kwargs):
        descriptions = []
        report_type_id = kwargs['report_type']
        if report_type_id != '':
            calendar_report_id = self.object.calendar_report_id
            _filter = CalendarReportDescription.objects.filter(
                report_type_id=report_type_id, calendar_report_id=calendar_report_id
            )
            search = kwargs.get('search')
            if search:
                _filter = _filter.filter(name__icontains=search)
            for description in _filter.values('id', 'name'):
                descriptions.append({'id': description['id'], 'text': description['name']})

        return JsonResponse({'results': descriptions})

    def form_valid(self, form):
        form.save()
        return self.command_response('reload')


class CalendarDataSetDuplicateModal(Modal):
    def modal_content(self):
        return 'Are you sure you want to duplicate this lane?'

    def get_modal_buttons(self):
        return [modal_button_method('Confirm', 'duplicate'), modal_button('Cancel', 'close', 'btn-secondary')]

    def button_duplicate(self, **_kwargs):
        calendar_report_data_set = get_object_or_404(CalendarReportDataSet, id=self.slug['pk'])
        calendar_report_data_set.pk = None
        calendar_report_data_set.name = f'Copy of {calendar_report_data_set.name}'
        calendar_report_data_set.order = None
        calendar_report_data_set.save()
        return self.command_response('reload')


class CalendarDescriptionModal(DataMergeUtils, QueryBuilderModalBase):
    template_name = 'advanced_report_builder/calendar/description_modal.html'
    size = 'xl'
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = CalendarReportDescription

    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name', 'report_type', 'description']

    def form_setup(self, form, *_args, **_kwargs):
        form.fields['description'].widget = DataMergeWidget()

        return ('name', 'report_type', 'description')

    def ajax_get_description_data_merge_menu(self, **kwargs):
        field_auto_id = kwargs['field_auto_id']
        menus = []
        if kwargs['report_type'] != '':
            report_builder_class, base_model = self.get_report_builder_class(report_type_id=kwargs['report_type'])

            self.get_menu_fields(base_model=base_model, report_builder_class=report_builder_class, menus=menus)

            menus = [{'code': 'data_merge', 'text': 'Data Merge', 'menu': menus}]

        return self.command_response(f'build_data_merge_menu_{field_auto_id}', data=json.dumps(menus))

    def form_valid(self, form):
        form.save()
        return self.command_response('reload')


class CalendarDescriptionDuplicateModal(Modal):
    def modal_content(self):
        return 'Are you sure you want to duplicate this description?'

    def get_modal_buttons(self):
        return [modal_button_method('Confirm', 'duplicate'), modal_button('Cancel', 'close', 'btn-secondary')]

    def button_duplicate(self, **_kwargs):
        calendar_report_description = get_object_or_404(CalendarReportDescription, id=self.slug['pk'])
        calendar_report_description.pk = None
        calendar_report_description.name = f'Copy of {calendar_report_description.name}'
        calendar_report_description.save()
        return self.command_response('reload')



# Generated by Django 5.0.7 on 2025-06-17 13:49

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0016_calendarreport_dashboardreport_options_and_more'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='barchartreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='calendarreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='customreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='dashboard',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='funnelchartreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='kanbanreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='linechartreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='piechartreport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='singlevaluereport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='tablereport',
            options={'get_latest_by': 'created'},
        ),
        migrations.AlterModelOptions(
            name='target',
            options={'get_latest_by': 'created'},
        ),
    ]



from django.forms.widgets import Input


class DataMergeWidget(Input):
    template_name = 'advanced_report_builder/data_merge/data_merge.html'
    input_type = 'textarea'
    crispy_kwargs = {
        'label_class': 'col-3 col-form-label-sm',
        'field_class': 'col-12 input-group-sm',
    }

    def __init__(self, data_merge_data=None, height=150):
        self.height = height
        super().__init__()
        self.data_merge_data = data_merge_data

    def get_context(self, name, value, attrs=None):
        context = super().get_context(name, value, attrs)
        if context['widget']['value'] is None:
            context['widget']['value'] = ''
        if self.data_merge_data is None:
            context['data_merge'] = None
        else:
            context['data_merge'] = self.data_merge_data.build_menus()

        context['disabled'] = attrs is not None and attrs.get('disabled', False)
        context['height'] = self.height
        return context



# Generated by Django 3.2.7 on 2022-02-09 16:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('advanced_report_builder', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='singlevaluereport',
            name='breakdown_fields',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='singlevaluereport',
            name='show_breakdown',
            field=models.BooleanField(default=False),
        ),
    ]



from django.views.generic import TemplateView
from django_menus.menu import MenuItem
from django_modals.modals import ModelFormModal
from django_modals.processes import PERMISSION_OFF, PROCESS_EDIT_DELETE
from django_modals.widgets.select2 import Select2Multiple

from advanced_report_builder.filter_query import FilterQueryMixin
from advanced_report_builder.models import CustomReport, ReportType
from advanced_report_builder.utils import split_slug
from advanced_report_builder.views.query_modal.mixin import MultiQueryModalMixin
from advanced_report_builder.views.report import ReportBase


class CustomBaseView(ReportBase, FilterQueryMixin, TemplateView):
    report_type_slug = None

    def __init__(self, **kwargs):
        self.report = None
        self.slug = None
        self.enable_edit = None
        self.dashboard_report = None
        self.show_toolbar = False
        self.enable_queries = True
        self._report_type = None
        super().__init__(**kwargs)

    def dispatch(self, request, *args, **kwargs):
        self.report = kwargs.get('report')
        self.slug = split_slug(kwargs.get('slug'))
        self.enable_edit = kwargs.get('enable_edit')
        self.dashboard_report = kwargs.get('dashboard_report')
        if self.enable_edit or (self.dashboard_report and not self.dashboard_report.top) or not self.dashboard_report:
            self.show_toolbar = True
        return super().dispatch(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['report'] = self.report
        context['show_toolbar'] = self.show_toolbar
        context['title'] = self.get_title()
        return context

    def setup_menu(self):
        super().setup_menu()
        if not self.show_toolbar:
            return

        if self.dashboard_report and self.enable_edit:
            report_menu = self.pod_dashboard_edit_menu()
        elif self.dashboard_report and not self.enable_edit:
            report_menu = self.pod_dashboard_view_menu()
        else:
            report_menu = self.pod_report_menu()

        self.add_menu('button_menu', 'button_group').add_items(
            *report_menu,
            *self.queries_menu(report=self.report, dashboard_report=self.dashboard_report),
        )

    def pod_dashboard_edit_menu(self):
        return [
            MenuItem(
                f'advanced_report_builder:dashboard_report_modal,pk-{self.dashboard_report.id}',
                menu_display='Edit',
                font_awesome='fas fa-pencil-alt',
                css_classes=['btn-primary'],
            )
        ]

    # noinspection PyMethodMayBeStatic
    def pod_dashboard_view_menu(self):
        return []

    # noinspection PyMethodMayBeStatic
    def pod_report_menu(self):
        return []

    def get_title(self):
        if self.dashboard_report and self.dashboard_report.name_override:
            return self.dashboard_report.name_override
        else:
            return self.report.name

    def get_report_type(self):
        if self.report_type_slug is not None:
            if self._report_type is None:
                self._report_type = ReportType.objects.get(slug=self.report_type_slug)
            return self._report_type
        return None

    def get_default_query(self):
        report_type = self.get_report_type()
        if report_type is not None:
            base_model = report_type.content_type.model_class()
            return base_model.objects.all()
        return None

    def get_default_filter(self, query):
        return query

    def get_query_results(self):
        query = self.get_default_query()
        if query is None:
            return None
        report_query = self.get_report_query(report=self.report)
        if report_query:
            query = self.process_query_filters(query=query, search_filter_data=report_query.query)
            report_type = self.get_report_type()
            if report_type is not None:
                base_model = report_type.content_type.model_class()
                query = self.apply_order_by(
                    query=query,
                    report_query=report_query,
                    report_type=report_type,
                    base_model=base_model,
                )
            return query
        return self.get_default_filter(query)


class CustomModal(MultiQueryModalMixin, ModelFormModal):
    process = PROCESS_EDIT_DELETE
    permission_delete = PERMISSION_OFF
    model = CustomReport
    ajax_commands = ['datatable', 'button']

    widgets = {'report_tags': Select2Multiple}

    form_fields = ['name', 'report_tags', 'notes']

    def form_setup(self, form, *_args, **_kwargs):
        fields = [*self.form_fields]
        if self.object.id and 'report_type' in self.slug:
            self.add_extra_queries(form=form, fields=fields)
        return fields

    def get_report_type(self, **_kwargs):
        return self.slug['report_type']



from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.dates import MONTHS
from django_datatables.columns import DatatableColumn, ManyToManyColumn, NoHeadingColumn
from django_datatables.model_def import DatatableModel
from django_modals.model_fields.colour import ColourField
from time_stamped_model.models import TimeStampedModel

from advanced_report_builder.globals import (
    ANNOTATION_CHART_SCALE,
    ANNOTATION_CHOICE_COUNT,
    ANNOTATION_VALUE_CHOICES,
    ANNOTATIONS_CHOICES,
    CALENDAR_VIEW_TYPE_CHOICES,
    CALENDAR_VIEW_TYPE_DAY,
    CALENDAR_VIEW_TYPE_GRID_WEEK,
    CALENDAR_VIEW_TYPE_LIST_WEEK,
    CALENDAR_VIEW_TYPE_MONTH,
    CALENDAR_VIEW_TYPE_YEAR,
    DISPLAY_OPTION_2_PER_ROW,
    DISPLAY_OPTION_CHOICES,
    DISPLAY_OPTION_CLASSES,
    DISPLAY_OPTION_NONE,
)
from advanced_report_builder.signals import model_report_save


class Target(TimeStampedModel):
    TARGET_TYPE_COUNT = 1
    TARGET_TYPE_MONEY = 2
    TARGET_TYPE_PERCENTAGE = 3

    TARGET_TYPE_CHOICES = (
        (TARGET_TYPE_COUNT, 'Count'),
        (TARGET_TYPE_MONEY, 'Money'),
        (TARGET_TYPE_PERCENTAGE, 'Percentage'),
    )

    slug = models.SlugField(unique=True)
    name = models.CharField(max_length=64)
    target_type = models.PositiveSmallIntegerField(choices=TARGET_TYPE_CHOICES)
    colour = ColourField(null=True, blank=True, help_text='The colour when it gets displayed on a report')
    default_value = models.IntegerField(blank=True, null=True)
    default_percentage = models.FloatField(blank=True, null=True)
    overridden = models.BooleanField(default=False)
    override_data = models.JSONField(null=True, blank=True)

    def get_value(self):
        if self.target_type == self.TARGET_TYPE_PERCENTAGE:
            return self.default_percentage
        return self.default_value

    def __str__(self):
        return self.name

    def get_override_data(self):
        override_data = self.override_data if self.override_data else {}

        ordered_value_dict = {}

        for year, data in override_data.items():
            if not ordered_value_dict.get(year):
                ordered_value_dict[year] = {}
            for month in MONTHS.values():
                if data.get(str(month)):
                    ordered_value_dict[year][month] = data[month]
                else:
                    ordered_value_dict[year][month] = self.default_value

        return ordered_value_dict

    def add_year(self, year):
        override_data = self.override_data if self.override_data else {}
        if not override_data.get(str(year)):
            override_data[str(year)] = {}
        temp = {}
        # Create a temp dict that has the keys as integers.
        for key, value in override_data.items():
            temp[int(key)] = value
        sorted_dict = {}
        for key in sorted(temp):
            sorted_dict[str(key)] = temp[key]
        if sorted_dict:
            self.override_data = sorted_dict
            self.save()

    def save(self, *args, **kwargs):
        """
        Overrides the save so that a new slug is generated
        :param args:
        :param kwargs:
        """
        self.make_new_slug()
        super().save(*args, **kwargs)


class ReportTag(TimeStampedModel):
    name = models.CharField(max_length=128, unique=True)
    slug = models.SlugField(unique=True)
    order = models.PositiveSmallIntegerField()

    class Meta:
        ordering = ('order',)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        self.make_new_slug(allow_dashes=False, on_edit=True)
        self.set_order_field()
        return super().save(*args, **kwargs)


class ReportType(TimeStampedModel):
    name = models.CharField(max_length=200)
    slug = models.SlugField(unique=True)
    content_type = models.ForeignKey(ContentType, null=False, blank=False, on_delete=models.PROTECT)
    report_builder_class_name = models.CharField(max_length=200)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def save(self, *args, **kwargs):
        self.make_new_slug(allow_dashes=False, on_edit=True)
        return super().save(*args, **kwargs)


class Report(TimeStampedModel):
    name = models.CharField(max_length=200)
    slug = models.SlugField(unique=True)
    slug_alias = models.SlugField(blank=True, null=True)  # used if the slug changes
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    instance_type = models.CharField(null=True, max_length=255)
    report_tags = models.ManyToManyField(ReportTag, blank=True)
    notes = models.TextField(null=True, blank=True)
    version = models.PositiveSmallIntegerField(default=0)
    user_created = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='report_user_created_set',
    )
    user_updated = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='report_user_updated_set',
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ['name']

    def get_report(self):
        return getattr(self, self.instance_type)

    def get_title(self):
        return self.name

    def save(self, *args, **kwargs):
        current_user = getattr(self, '_current_user', None)
        if current_user is not None and current_user.is_authenticated:
            self.user_updated = current_user
            if self.user_created is None:
                self.user_created = current_user
        is_new = self.pk is None
        if self.version is not None:
            self.version += 1
        slug_alias = self.slug
        self.make_new_slug(obj=Report, allow_dashes=False, on_edit=True)
        if slug_alias != self.slug:
            self.slug_alias = slug_alias
        if self.instance_type is None:
            self.instance_type = self._meta.label_lower.split('.')[1]
        result = super().save(*args, **kwargs)
        model_report_save.send(sender=self.__class__, instance=self, created=is_new, user=current_user)
        return result

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    class Datatable(DatatableModel):
        class OutputType(DatatableColumn):
            output_types = {
                'tablereport': 'Table',
                'singlevaluereport': 'Single Value',
                'barchartreport': 'Bar Chart',
                'linechartreport': 'Line Chart',
                'piechartreport': 'Pie Chart',
                'funnelchartreport': 'Funnel Chart',
                'kanbanreport': 'Kanban',
                'customreport': 'Custom',
                'calendarreport': 'Calendar',
            }

            def col_setup(self):
                self.field = ['instance_type', 'customreport__output_type']

            # noinspection PyMethodMayBeStatic
            def row_result(self, data, _page_data):
                instance_type = data[self.model_path + 'instance_type']
                if instance_type == 'customreport':
                    output_type = data[self.model_path + 'customreport__output_type']
                    if output_type:
                        return output_type
                return self.output_types.get(instance_type, '')

        class OutputTypeIcon(NoHeadingColumn):
            output_types = {
                'tablereport': '<i class="fas fa-table"></i>',
                'singlevaluereport': '<i class="fas fa-box-open"></i>',
                'barchartreport': '<i class="fas fa-chart-bar"></i>',
                'linechartreport': '<i class="fas fa-chart-line"></i>',
                'piechartreport': '<i class="fas fa-chart-pie"></i>',
                'funnelchartreport': '<i class="fas fa-filter"></i>',
                'kanbanreport': '<i class="fas fa-chart-bar fa-flip-vertical"></i>',
                'customreport': '<i class="fas fa-file"></i>',
                'calendarreport': '<i class="fas fa-calendar"></i>',
            }

            def col_setup(self):
                self.field = ['instance_type']

            # noinspection PyMethodMayBeStatic
            def row_result(self, data, _page_data):
                instance_type = data[self.model_path + 'instance_type']
                return self.output_types.get(instance_type, '')

        report_tags_badge = ManyToManyColumn(
            field='report_tags__name',
            html='<span class="badge badge-primary"> %1% </span>',
            title='Tags',
        )


class ReportQuery(TimeStampedModel):
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    name = models.CharField(max_length=256, default='Standard')
    query = models.JSONField(null=True, blank=True)
    extra_query = models.JSONField(null=True, blank=True)  # used for single value Numerator
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'report': self.report})
        result = super().save(*args, **kwargs)
        self.report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.report.save()
        return result

    class Meta:
        ordering = ['order']
        verbose_name_plural = 'Report queries'

    def __str__(self):
        return self.name


class ReportQueryOrder(TimeStampedModel):
    report_query = models.ForeignKey(ReportQuery, on_delete=models.CASCADE)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    order = models.PositiveSmallIntegerField()

    class Meta:
        ordering = ['order']

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'report_query': self.report_query})
        return super().save(*args, **kwargs)


class TableReport(Report):
    table_fields = models.JSONField(null=True, blank=True)
    has_clickable_rows = models.BooleanField(default=False)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    pivot_fields = models.JSONField(null=True, blank=True)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    page_length = models.PositiveSmallIntegerField(
        choices=(
            (10, '10'),
            (25, '25'),
            (50, '50'),
            (100, '100'),
            (150, '150'),
            (200, '200'),
        ),
        default=100,
    )


class SingleValueReport(Report):
    SINGLE_VALUE_TYPE_COUNT = 1
    SINGLE_VALUE_TYPE_SUM = 2
    SINGLE_VALUE_TYPE_COUNT_AND_SUM = 3
    SINGLE_VALUE_TYPE_PERCENT = 4
    SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT = 5
    SINGLE_VALUE_TYPE_AVERAGE_SUM_FROM_COUNT = 6
    SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME = 7

    SINGLE_VALUE_TYPE_CHOICES = (
        (SINGLE_VALUE_TYPE_COUNT, 'Count'),
        (SINGLE_VALUE_TYPE_SUM, 'Sum'),
        (SINGLE_VALUE_TYPE_COUNT_AND_SUM, 'Count & Sum'),
        (SINGLE_VALUE_TYPE_PERCENT, 'Percent'),
        (SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT, 'Percent from Count'),
        (SINGLE_VALUE_TYPE_AVERAGE_SUM_FROM_COUNT, 'Average Sum from Count'),
        (SINGLE_VALUE_TYPE_AVERAGE_SUM_OVER_TIME, 'Average Sum over Time'),
    )

    tile_colour = ColourField(blank=True, null=True)
    field = models.CharField(max_length=200, blank=True, null=True)  # denominator
    numerator = models.CharField(max_length=200, blank=True, null=True)
    single_value_type = models.PositiveSmallIntegerField(
        choices=SINGLE_VALUE_TYPE_CHOICES, default=SINGLE_VALUE_TYPE_COUNT
    )
    prefix = models.CharField(max_length=64, blank=True, null=True)
    decimal_places = models.IntegerField(default=0)
    show_breakdown = models.BooleanField(default=False)
    breakdown_fields = models.JSONField(null=True, blank=True)

    average_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES, blank=True, null=True)
    average_start_period = models.PositiveSmallIntegerField(blank=True, null=True)
    average_end_period = models.PositiveSmallIntegerField(blank=True, null=True)

    def is_percentage(self):
        return self.single_value_type in [
            self.SINGLE_VALUE_TYPE_PERCENT,
            self.SINGLE_VALUE_TYPE_PERCENT_FROM_COUNT,
        ]


class BarChartReport(Report):
    BAR_CHART_ORIENTATION_VERTICAL = 1
    BAR_CHART_ORIENTATION_HORIZONTAL = 2

    BAR_CHART_ORIENTATION_CHOICES = (
        (BAR_CHART_ORIENTATION_VERTICAL, 'Vertical'),
        (BAR_CHART_ORIENTATION_HORIZONTAL, 'Horizontal'),
    )

    DATE_FIELD_SINGLE = 1
    DATE_FIELD_RANGE = 2

    DATE_FIELD_CHOICES = ((DATE_FIELD_SINGLE, 'Single'), (DATE_FIELD_RANGE, 'Range'))

    axis_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES)

    date_field_type = models.PositiveSmallIntegerField(choices=DATE_FIELD_CHOICES, default=DATE_FIELD_SINGLE)

    date_field = models.CharField(max_length=200)
    end_date_field = models.CharField(max_length=200, blank=True, null=True)

    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    x_label = models.CharField(max_length=200, blank=True, null=True)
    y_label = models.CharField(max_length=200, blank=True, null=True)

    bar_chart_orientation = models.PositiveSmallIntegerField(
        choices=BAR_CHART_ORIENTATION_CHOICES, default=BAR_CHART_ORIENTATION_VERTICAL
    )
    stacked = models.BooleanField(default=False)
    show_totals = models.BooleanField(default=False)
    show_blank_dates = models.BooleanField(default=True)

    show_breakdown = models.BooleanField(default=False)
    breakdown_fields = models.JSONField(null=True, blank=True)

    def is_orientation_vertical(self):
        return self.bar_chart_orientation == self.BAR_CHART_ORIENTATION_VERTICAL

    def get_chart_scale(self):
        return ANNOTATION_CHART_SCALE[self.axis_scale]


class LineChartReport(Report):
    axis_scale = models.PositiveSmallIntegerField(choices=ANNOTATION_VALUE_CHOICES)
    date_field = models.CharField(max_length=200)
    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    x_label = models.CharField(max_length=200, blank=True, null=True)
    y_label = models.CharField(max_length=200, blank=True, null=True)
    show_totals = models.BooleanField(default=False)

    has_targets = models.BooleanField(default=False)
    targets = models.ManyToManyField(Target, blank=True)

    def get_chart_scale(self):
        return ANNOTATION_CHART_SCALE[self.axis_scale]


class PieChartReport(Report):
    PIE_CHART_STYLE_PIE = 1
    PIE_CHART_STYLE_DOUGHNUT = 2

    PIE_CHART_STYLE_CHOICES = (
        (PIE_CHART_STYLE_PIE, 'Pie'),
        (PIE_CHART_STYLE_DOUGHNUT, 'Doughnut'),
    )

    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)
    style = models.PositiveSmallIntegerField(choices=PIE_CHART_STYLE_CHOICES, default=PIE_CHART_STYLE_PIE)

    def is_pie_chart(self):
        return self.style == self.PIE_CHART_STYLE_PIE


class FunnelChartReport(Report):
    axis_value_type = models.PositiveSmallIntegerField(
        choices=ANNOTATIONS_CHOICES,
        default=ANNOTATION_CHOICE_COUNT,
        null=True,
        blank=True,
    )
    fields = models.JSONField(null=True, blank=True)


class KanbanReport(Report):
    pass


class KanbanReportDescription(TimeStampedModel):
    kanban_report = models.ForeignKey(KanbanReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    description = models.TextField(blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'kanban_report': self.kanban_report})
        result = super().save(*args, **kwargs)
        self.kanban_report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.kanban_report.save()
        return result

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('order',)


class CalendarReport(Report):
    VIEW_TYPE_CODES = {
        CALENDAR_VIEW_TYPE_MONTH: 'dayGridMonth',
        CALENDAR_VIEW_TYPE_GRID_WEEK: 'timeGridWeek',
        CALENDAR_VIEW_TYPE_LIST_WEEK: 'listWeek',
        CALENDAR_VIEW_TYPE_DAY: 'timeGridDay',
        CALENDAR_VIEW_TYPE_YEAR: 'year',
    }

    height = models.PositiveSmallIntegerField(default=600)
    view_type = models.PositiveSmallIntegerField(
        choices=CALENDAR_VIEW_TYPE_CHOICES,
        default=CALENDAR_VIEW_TYPE_MONTH,
    )

    def get_view_type_for_calendar(self, view_type=None):
        if view_type is None:
            view_type = self.view_type
        return self.VIEW_TYPE_CODES.get(view_type)


class CalendarReportDescription(TimeStampedModel):
    calendar_report = models.ForeignKey(CalendarReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    description = models.TextField(blank=True, null=True)
    order = models.PositiveSmallIntegerField()

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'calendar_report': self.calendar_report})
        return super().save(*args, **kwargs)

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('order',)


class CalendarReportDataSet(TimeStampedModel):
    DISPLAY_TYPE_NAME_AND_DESCRIPTION = 1
    DISPLAY_TYPE_DESCRIPTION_ONLY = 2
    DISPLAY_TYPE_HEADING_ONLY = 3

    DISPLAY_TYPE_CHOICES = (
        (DISPLAY_TYPE_NAME_AND_DESCRIPTION, 'Name and Description'),
        (DISPLAY_TYPE_DESCRIPTION_ONLY, 'Description Only'),
        (DISPLAY_TYPE_HEADING_ONLY, 'Heading Only'),
    )

    END_DATE_TYPE_FIELD = 1
    END_DATE_TYPE_DURATION_FIELD = 2
    END_DATE_TYPE_DURATION_FIXED = 3

    END_DATE_TYPE_CHOICES = (
        (END_DATE_TYPE_FIELD, 'Field'),
        (END_DATE_TYPE_DURATION_FIELD, 'Duration Field'),
        (END_DATE_TYPE_DURATION_FIXED, 'Duration Fixed'),
    )

    calendar_report = models.ForeignKey(CalendarReport, on_delete=models.CASCADE)
    order = models.PositiveSmallIntegerField()
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    heading_field = models.CharField(max_length=200, blank=True, null=True)
    name = models.CharField(max_length=200)
    display_type = models.PositiveSmallIntegerField(
        choices=DISPLAY_TYPE_CHOICES,
        default=DISPLAY_TYPE_NAME_AND_DESCRIPTION,
    )
    start_date_field = models.CharField(max_length=200, blank=True, null=True)
    end_date_type = models.PositiveSmallIntegerField(
        choices=END_DATE_TYPE_CHOICES,
        default=END_DATE_TYPE_FIELD,
    )
    end_date_field = models.CharField(max_length=200, blank=True, null=True)
    end_duration_field = models.CharField(max_length=200, blank=True, null=True)
    end_duration = models.PositiveSmallIntegerField(blank=True, null=True)
    background_colour_field = models.CharField(max_length=200, blank=True, null=True)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    calendar_report_description = models.ForeignKey(
        CalendarReportDescription, null=True, blank=False, on_delete=models.CASCADE
    )
    query_data = models.JSONField(null=True, blank=True)

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'calendar_report': self.calendar_report})
        return super().save(*args, **kwargs)

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    class Meta:
        ordering = ('order',)


class KanbanReportLane(TimeStampedModel):
    MULTIPLE_TYPE_NA = 0
    MULTIPLE_TYPE_DAILY = 1
    MULTIPLE_TYPE_DAILY_WITHIN = 2
    MULTIPLE_TYPE_WEEKLY = 3
    MULTIPLE_TYPE_WEEKLY_WITHIN = 4
    MULTIPLE_TYPE_MONTHLY = 5
    MULTIPLE_TYPE_MONTHLY_WITHIN = 6

    MULTIPLE_TYPE_CHOICES = [
        (MULTIPLE_TYPE_NA, 'N/A'),
        (MULTIPLE_TYPE_DAILY, 'Daily (single date)'),
        (MULTIPLE_TYPE_DAILY_WITHIN, 'Daily (within two date)'),
        (MULTIPLE_TYPE_WEEKLY, 'Weekly (single date)'),
        (MULTIPLE_TYPE_WEEKLY_WITHIN, 'Weekly (within two date)'),
        (MULTIPLE_TYPE_MONTHLY, 'Monthly (single date)'),
        (MULTIPLE_TYPE_MONTHLY_WITHIN, 'Monthly (within two date)'),
    ]

    kanban_report = models.ForeignKey(KanbanReport, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    order = models.PositiveSmallIntegerField()
    report_type = models.ForeignKey(ReportType, null=True, blank=False, on_delete=models.PROTECT)
    heading_field = models.CharField(max_length=200, blank=True, null=True)
    link_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_field = models.CharField(max_length=200, blank=True, null=True)
    order_by_ascending = models.BooleanField(default=True)
    kanban_report_description = models.ForeignKey(
        KanbanReportDescription, null=True, blank=False, on_delete=models.CASCADE
    )

    multiple_type = models.PositiveIntegerField(choices=MULTIPLE_TYPE_CHOICES, default=MULTIPLE_TYPE_NA)
    multiple_type_label = models.CharField(max_length=200, blank=True, null=True)
    multiple_type_date_field = models.CharField(max_length=200, blank=True, null=True)
    multiple_type_end_date_field = models.CharField(max_length=200, blank=True, null=True)

    # this could be choice field from RANGE_TYPE_CHOICES however if one adds a new one it creates a new migration!
    multiple_start_period = models.PositiveSmallIntegerField(blank=True, null=True)
    multiple_end_period = models.PositiveSmallIntegerField(blank=True, null=True)
    background_colour_field = models.CharField(max_length=200, blank=True, null=True)
    heading_colour_field = models.CharField(max_length=200, blank=True, null=True)

    query_data = models.JSONField(null=True, blank=True)

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'kanban_report': self.kanban_report})
        result = super().save(*args, **kwargs)
        self.kanban_report._current_user = getattr(self, '_current_user', None)  # this is to update the users
        self.kanban_report.save()
        return result

    def get_base_model(self):
        return self.report_type.content_type.model_class()

    class Meta:
        ordering = ('order',)


class CustomReport(Report):
    output_type = models.CharField(max_length=200, blank=True, null=True)
    view_name = models.CharField(max_length=200)
    settings = models.JSONField(null=True, blank=True)


class Dashboard(TimeStampedModel):
    slug = models.SlugField(unique=True)
    slug_alias = models.SlugField(blank=True, null=True)  # used if the slug changes
    name = models.CharField(max_length=200)
    display_option = models.PositiveIntegerField(choices=DISPLAY_OPTION_CHOICES[1:], default=DISPLAY_OPTION_2_PER_ROW)

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        slug_alias = self.slug
        self.make_new_slug(allow_dashes=False, on_edit=True)
        if slug_alias != self.slug:
            self.slug_alias = slug_alias
        return super().save(*args, **kwargs)


class DashboardReport(TimeStampedModel):
    dashboard = models.ForeignKey(Dashboard, on_delete=models.CASCADE)
    order = models.PositiveSmallIntegerField()
    report = models.ForeignKey(Report, on_delete=models.CASCADE)
    top = models.BooleanField(default=False)
    name_override = models.CharField(max_length=200, blank=True, null=True)
    display_option = models.PositiveIntegerField(choices=DISPLAY_OPTION_CHOICES, default=DISPLAY_OPTION_NONE)
    show_versions = models.BooleanField(default=True)
    report_query = models.ForeignKey(ReportQuery, blank=True, null=True, on_delete=models.CASCADE)
    options = models.JSONField(null=True, blank=True)

    def get_class(self, extra_class_name):
        if self.display_option != DISPLAY_OPTION_NONE:
            class_names = DISPLAY_OPTION_CLASSES.get(self.display_option)
        else:
            class_names = DISPLAY_OPTION_CLASSES.get(self.dashboard.display_option)

        if extra_class_name:
            class_names += f' {extra_class_name}'

        return class_names

    def save(self, *args, **kwargs):
        self.set_order_field(extra_filters={'dashboard': self.dashboard, 'top': self.top})
        return super().save(*args, **kwargs)

    class Meta:
        ordering = ['order']



from django.contrib.humanize.templatetags.humanize import intcomma
from django.contrib.postgres.aggregates import ArrayAgg, BoolAnd, BoolOr, StringAgg
from django.db.models import BooleanField, Count, Max, Min
from django.db.models.functions import Cast
from django_datatables.columns import (
    ColumnBase,
    ColumnLink,
    CurrencyColumn,
    CurrencyPenceColumn,
    ManyToManyColumn,
    NoHeadingColumn,
)
from django_datatables.helpers import DUMMY_ID, get_url, render_replace
from django_datatables.model_def import DatatableModel

from advanced_report_builder.globals import (
    DATE_FORMAT_TYPE_DD_MM_YY_SLASH,
    DATE_FORMAT_TYPES_DJANGO_FORMAT,
    REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_AND,
    REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_ARRAY,
    REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR,
    REVERSE_FOREIGN_KEY_ANNOTATION_DATE_ARRAY,
    REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MAX,
    REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MIN,
    REVERSE_FOREIGN_KEY_DELIMITER_COMMA,
    REVERSE_FOREIGN_KEY_DELIMITER_VALUES,
)


class ReportBuilderDateColumn(ColumnBase):
    def __init__(self, *, date_format=None, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        if date_format:
            self.date_format = date_format
        else:
            self.date_format = '%d/%m/%Y'

    def row_result(self, data, _page_data):
        try:
            date = data[self.field].strftime(self.date_format)
            return date
        except AttributeError:
            return ''


class ReportBuilderNumberColumn(ColumnBase):
    def __init__(self, *, decimal_places=0, trim_zeros=True, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.decimal_places = f'{{:.{decimal_places}f}}'
        self.trim_zeros = trim_zeros

    def row_result(self, data, _page_data):
        number = data.get(self.field)
        if number is None:
            return ''
        else:
            number = self.decimal_places.format(number)
        if self.trim_zeros and '.' in number:
            return number.rstrip('0').rstrip('.')
        else:
            return number


class ReportBuilderCurrencyPenceColumn(CurrencyPenceColumn):
    def row_result(self, data, _page_data):
        try:
            return intcomma(f'{data[self.field] / 100.0:.2f}')
        except (KeyError, TypeError):
            return '0.00'


class ReportBuilderCurrencyColumn(CurrencyColumn):
    def row_result(self, data, _page_data):
        try:
            return intcomma(f'{data[self.field]:.2f}')
        except (KeyError, TypeError):
            return '0.00'


class ArrowColumn(NoHeadingColumn):
    def __init__(self, **kwargs):
        if not self.initialise(locals()):
            return
        kwargs['render'] = [{'html': '<i class="fa fa-angle-right fa-2x"></i>', 'function': 'Html'}]
        kwargs['width'] = '10px'
        super().__init__(**kwargs)


class ColourColumn(ColumnBase):
    def __init__(self, **kwargs):
        if not self.initialise(locals()):
            return
        kwargs['render'] = [
            render_replace(
                html='<span style="display: inline-block; width: 60px; height: 15px;'
                ' background-color: #%1%; vertical-align: middle;"></span>',
                column=kwargs['column_name'],
            )
        ]
        super().__init__(**kwargs)


class FilterForeignKeyColumn(ColumnBase):
    def get_query_options(self):
        values = self.model.objects.distinct(self.field).order_by(self.field).values_list(self.field, flat=True)
        return {v: v for v in values if v}


class ReverseForeignKeyStrColumn(ColumnBase):
    def __init__(self, field_name, report_builder_class_name, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.field_name = field_name
        self.report_builder_class_name = report_builder_class_name

    def setup_annotations(self, delimiter_type=None, sub_filter=None, field_name=None):
        if field_name is None:
            field_name = self.field_name
        delimiter = REVERSE_FOREIGN_KEY_DELIMITER_VALUES[delimiter_type]

        self.annotations = {
            field_name: StringAgg(self.field_name, delimiter=delimiter, distinct=True, filter=sub_filter)
        }


class ReverseForeignKeyBoolColumn(ColumnBase):
    def __init__(self, field_name, report_builder_class_name, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.field_name = field_name
        self.report_builder_class_name = report_builder_class_name

    def setup_annotations(self, annotations_type, sub_filter=None, field_name=None):
        if field_name is None:
            field_name = self.field_name

        if annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_XOR:
            self.annotations = {field_name: BoolOr(Cast(self.field_name, BooleanField()), filter=sub_filter)}
        elif annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_AND:
            self.annotations = {field_name: BoolAnd(Cast(self.field_name, BooleanField()), filter=sub_filter)}
        elif annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_BOOLEAN_ARRAY:
            self.annotations = {field_name: ArrayAgg(self.field_name, distinct=True, filter=sub_filter)}


class ReverseForeignKeyChoiceColumn(ColumnBase):
    def __init__(self, field_name, report_builder_class_name, **kwargs):
        if not self.initialise(locals()):
            return
        self.field_name = field_name
        self.choices = None
        self.report_builder_class_name = report_builder_class_name
        self.delimiter_type = REVERSE_FOREIGN_KEY_DELIMITER_COMMA
        super().__init__(**kwargs)

    def setup_results(self, request, all_results):
        _, django_field, _ = DatatableModel.get_setup_data(self.model, self.field_name)
        self.choices = {c[0]: c[1] for c in django_field.choices}

    def row_result(self, data, _page_data):
        results = []
        for x in data.get(self.field):
            results.append(self.choices.get(x, ''))
        delimiter = REVERSE_FOREIGN_KEY_DELIMITER_VALUES[self.delimiter_type]
        return delimiter.join(results)

    def setup_annotations(self, delimiter_type=None, sub_filter=None, field_name=None):
        if field_name is None:
            field_name = self.field_name
        self.delimiter_type = delimiter_type
        self.annotations = {field_name: ArrayAgg(self.field_name, distinct=True, filter=sub_filter)}


class ReverseForeignKeyDateColumn(ColumnBase):
    def __init__(self, field_name, report_builder_class_name, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(**kwargs)
        self.field_name = field_name
        self.report_builder_class_name = report_builder_class_name
        self.delimiter_type = REVERSE_FOREIGN_KEY_DELIMITER_COMMA
        self.date_format_type = DATE_FORMAT_TYPE_DD_MM_YY_SLASH
        self.annotations_type = REVERSE_FOREIGN_KEY_ANNOTATION_DATE_ARRAY

    def setup_annotations(
        self,
        delimiter_type=None,
        date_format_type=None,
        annotations_type=None,
        sub_filter=None,
        field_name=None,
    ):
        if field_name is None:
            field_name = self.field_name
        self.delimiter_type = delimiter_type
        self.date_format_type = date_format_type
        self.annotations_type = annotations_type

        if annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_DATE_ARRAY:
            self.annotations = {field_name: ArrayAgg(self.field_name, distinct=True, filter=sub_filter)}
        elif annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MIN:
            self.annotations = {field_name: Min(self.field_name, sub_filter=sub_filter)}

        elif annotations_type == REVERSE_FOREIGN_KEY_ANNOTATION_DATE_MAX:
            self.annotations = {field_name: Max(self.field_name, sub_filter=sub_filter)}

    def row_result(self, data, _page_data):
        field = self.field
        if isinstance(field, list):
            field = field[-1]
        try:
            date_formate_type_str = DATE_FORMAT_TYPES_DJANGO_FORMAT[self.date_format_type]
            results = []
            date_values = data[field]
            if date_values is None:
                return ''
            if isinstance(date_values, list):
                for value in date_values:
                    results.append(value.strftime(date_formate_type_str))
                delimiter = REVERSE_FOREIGN_KEY_DELIMITER_VALUES[self.delimiter_type]
                return delimiter.join(results)
            else:
                return date_values.strftime(date_formate_type_str)
        except AttributeError:
            return ''


class ReportBuilderColumnLink(ColumnLink):
    """Sometimes you may want to have a report where the links don't work.
    This is used for when you have a wall board"""

    @property
    def url(self):
        return self._url

    def setup_link(self, link_css, link_html):
        if self.enable_links():
            link_css = f' class="{link_css}"' if link_css else ''
            link = f'<a{link_css} href="{self.url}">{{}}</a>'
        else:
            link = '{}'
        if isinstance(self.field, list | tuple):
            self.options['render'] = [
                render_replace(column=self.column_name + ':0', html=link.format(link_html), var='999999'),
                render_replace(column=self.column_name + ':1'),
            ]
        elif self.var not in link_html:
            self.options['render'] = [
                render_replace(column=self.link_ref_column, html=link.format(link_html), var='999999')
            ]
        else:
            self.options['render'] = [
                render_replace(column=self.column_name, html=link.format(link_html), var=self.var),
                render_replace(column=self.link_ref_column, var='999999'),
            ]

    @url.setter
    def url(self, url_name):
        if self.enable_links():
            self._url = get_url(url_name)
        else:
            self._url = f'#?{DUMMY_ID}'

    def enable_links(self):
        return (
            not self.table
            or self.table.view is None
            or self.table.view.kwargs.get('enable_links')
            or getattr(self.table.view, 'enable_links', False)
        )


class RecordCountColumn(ColumnBase):
    def __init__(self, field=None, **kwargs):
        if 'annotations' not in kwargs:
            kwargs['annotations'] = {'record_count': Count(1)}
        if 'column_name' not in kwargs:
            kwargs['column_name'] = 'record_count'
        super().__init__(field, **kwargs)


class ReportBuilderManyToManyColumn(ManyToManyColumn):
    def __init__(self, *arg, **kwargs):
        if not self.initialise(locals()):
            return
        super().__init__(*arg, **kwargs)
        if self.model_path is not None:
            self.field = 'id'

    def row_result(self, data_dict, page_results):
        return page_results['m2m' + self.column_name].get(data_dict[self.model_path + 'id'], self.blank)



import setuptools

with open('README.md') as fh:
    long_description = fh.read()

setuptools.setup(
    name='django-advanced-report-builder',
    version='1.0.6',
    author='Thomas Turner',
    description='Django app that allows you to build reports from modals',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/django-advance-utils/django-advanced-report-builder',
    include_package_data=True,
    packages=['advanced_report_builder'],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    install_requires=[
        'Django >= 3.2',
        'django-filtered-datatables >= 0.0.21',
        'django-ajax-helpers >= 0.0.20',
        'django-nested-modals >= 0.0.21',
        'time-stamped-model >= 0.2.3',
        'date-offset >= 0.0.2',
    ],
)



from django.apps import apps
from django_datatables.columns import ColumnNameError
from django_datatables.datatables import ColumnInitialisor

from advanced_report_builder.column_types import (
    COLOUR_COLUMNS,
    DATE_FIELDS,
    LINK_COLUMNS,
    NUMBER_FIELDS,
)
from advanced_report_builder.exceptions import ReportError
from advanced_report_builder.utils import get_report_builder_class


class ReportBuilderFieldUtils:
    column_initialisor_cls = ColumnInitialisor

    def get_field_details(self, base_model, field, report_builder_class, table=None, field_attr=None):
        if field_attr is None:
            field_attr = {}
        if isinstance(field, str) and field in report_builder_class.field_classes:
            field = report_builder_class.field_classes[field]

        path = field
        original_column_initialisor = self.column_initialisor_cls(
            start_model=base_model, path=field, table=table, **field_attr
        )

        try:
            columns = original_column_initialisor.get_columns()
        except ColumnNameError as e:
            raise ReportError(e)
        django_field = original_column_initialisor.django_field
        col_type_override = None
        if columns:
            col_type_override = columns[0]
            if isinstance(field, str) and '__' in field and col_type_override.model_path == '':
                # I think there is a bug in datatables where the model path isn't always set! hence the following code
                field_parts = field.split('__')[:-1]
                model_path = '__'.join(field_parts) + '__'
                col_type_override.model_path = model_path
                col_type_override.field = col_type_override.field  # actions the setter
                path = model_path

        if django_field is None and columns:
            col_type_override = columns[0]
            if isinstance(col_type_override.field, str):
                if isinstance(field, str):
                    path_parts = field.split('__')[:-1]
                    path_parts.append(col_type_override.field.split('__')[-1])
                    path = '__'.join(path_parts)
                    column_initialisor = self.column_initialisor_cls(
                        start_model=base_model, path=path, table=table, **field_attr
                    )
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field
                else:
                    column_initialisor = self.column_initialisor_cls(
                        start_model=base_model,
                        path=col_type_override.field,
                        table=table,
                        **field_attr,
                    )
                    column_initialisor.get_columns()
                    django_field = column_initialisor.django_field

        return django_field, col_type_override, columns, path

    def get_field_display_value(
        self,
        field_type,
        fields_values,
        base_model,
        report_builder_class,
        selected_field_value,
        for_select2=False,
    ):
        if field_type == 'date':
            self._get_date_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
            )
        elif field_type == 'link':
            self._get_column_link_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
            )
        elif field_type == 'number':
            self._get_number_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
            )
        elif field_type == 'colour':
            self._get_colour_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
            )
        elif field_type == 'all':
            self._get_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
                for_select2=for_select2,
            )
        elif field_type == 'order' or field_type == 'django_order':
            self._get_fields(
                base_model=base_model,
                fields=fields_values,
                report_builder_class=report_builder_class,
                selected_field_id=selected_field_value,
                for_select2=for_select2,
                show_order_by_fields=True,
            )

    def _get_date_fields(
        self,
        base_model,
        fields,
        report_builder_class,
        selected_field_id=None,
        search_string=None,
    ):
        return self._get_fields(
            base_model=base_model,
            fields=fields,
            report_builder_class=report_builder_class,
            selected_field_id=selected_field_id,
            field_types=DATE_FIELDS,
            for_select2=True,
            allow_annotations_fields=False,
            search_string=search_string,
        )

    def _get_number_fields(
        self,
        base_model,
        fields,
        report_builder_class,
        selected_field_id=None,
        search_string=None,
    ):
        return self._get_fields(
            base_model=base_model,
            fields=fields,
            report_builder_class=report_builder_class,
            selected_field_id=selected_field_id,
            field_types=NUMBER_FIELDS,
            for_select2=True,
            search_string=search_string,
        )

    def _get_column_link_fields(
        self,
        base_model,
        fields,
        report_builder_class,
        selected_field_id=None,
        search_string=None,
    ):
        return self._get_fields(
            base_model=base_model,
            fields=fields,
            report_builder_class=report_builder_class,
            selected_field_id=selected_field_id,
            column_types=LINK_COLUMNS,
            for_select2=True,
            search_string=search_string,
            allow_annotations_fields=False,
        )

    def _get_colour_fields(
        self,
        base_model,
        fields,
        report_builder_class,
        selected_field_id=None,
        search_string=None,
    ):
        return self._get_fields(
            base_model=base_model,
            fields=fields,
            report_builder_class=report_builder_class,
            selected_field_id=selected_field_id,
            column_types=COLOUR_COLUMNS,
            for_select2=True,
            search_string=search_string,
            allow_annotations_fields=False,
        )

    def _get_fields(
        self,
        base_model,
        fields,
        report_builder_class,
        tables=None,
        prefix='',
        title_prefix='',
        title=None,
        colour=None,
        previous_base_model=None,
        selected_field_id=None,
        for_select2=False,
        pivot_fields=None,
        allow_annotations_fields=True,
        field_types=None,
        column_types=None,
        search_string=None,
        show_order_by_fields=False,
        extra_fields=None,
        must_have_django_field=False,
        allow_pivots=True,
        include_mathematical_columns=False,
    ):
        if title is None:
            title = report_builder_class.title
        if colour is None:
            colour = report_builder_class.colour

        if tables is not None:
            tables.append({'name': title, 'colour': colour})

        report_builder_class_fields = report_builder_class.fields

        if extra_fields:
            report_builder_class_fields += extra_fields

        for report_builder_field in report_builder_class_fields:
            if (
                not isinstance(report_builder_field, str)
                or report_builder_field not in report_builder_class.exclude_display_fields
                or (show_order_by_fields and report_builder_field in report_builder_class.order_by_fields)
            ):
                django_field, col_type_override, columns, _ = self.get_field_details(
                    base_model=base_model,
                    field=report_builder_field,
                    report_builder_class=report_builder_class,
                )
                if django_field is None and must_have_django_field:
                    continue
                for column in columns:
                    if (
                        (field_types is None and column_types is None)
                        or (field_types is not None and isinstance(django_field, field_types))
                        or (column_types is not None and isinstance(col_type_override, column_types))
                        or (allow_annotations_fields and column.annotations)
                    ):
                        full_id = prefix + column.column_name
                        if selected_field_id is None or selected_field_id == full_id:
                            if column.title == '':
                                full_title = title_prefix + col_type_override.title_from_name(column.column_name)
                            else:
                                full_title = title_prefix + column.title
                            if self._is_search_match(search_string=search_string, title=full_title):
                                if for_select2:
                                    fields.append({'id': full_id, 'text': full_title})
                                else:
                                    fields.append(
                                        {
                                            'field': full_id,
                                            'label': full_title,
                                            'colour': colour,
                                        }
                                    )

        if allow_pivots and not for_select2 and pivot_fields is not None:
            for pivot_code, pivot_field in report_builder_class.pivot_fields.items():
                full_id = prefix + pivot_code
                full_title = title_prefix + pivot_field['title']
                if self._is_search_match(search_string=search_string, title=full_title):
                    pivot_fields.append(
                        {
                            'field': full_id,
                            'label': title_prefix + pivot_field['title'],
                            'colour': colour,
                        }
                    )

        for include_field, include in report_builder_class.includes.items():
            app_label, model, report_builder_fields_str = include['model'].split('.')
            local_allow_pivots = allow_pivots
            if local_allow_pivots and not include.get('allow_pivots', True):
                local_allow_pivots = False

            new_model = apps.get_model(app_label, model)
            if new_model != previous_base_model:
                new_report_builder_class = get_report_builder_class(
                    model=new_model, class_name=report_builder_fields_str
                )
                self._get_fields(
                    base_model=new_model,
                    fields=fields,
                    report_builder_class=new_report_builder_class,
                    tables=tables,
                    prefix=f'{prefix}{include_field}__',
                    title_prefix=f'{title_prefix}{include["title"]} -> ',
                    title=include.get('title'),
                    colour=include.get('colour'),
                    previous_base_model=base_model,
                    selected_field_id=selected_field_id,
                    for_select2=for_select2,
                    pivot_fields=pivot_fields,
                    allow_annotations_fields=allow_annotations_fields,
                    field_types=field_types,
                    column_types=column_types,
                    search_string=search_string,
                    show_order_by_fields=show_order_by_fields,
                    allow_pivots=local_allow_pivots,
                )

        if include_mathematical_columns:
            fields.extend(
                [
                    {
                        'colour': '#D4AF37',
                        'field': 'rb_addition',
                        'label': 'Maths: Addition Field',
                    },
                    {
                        'colour': '#D4AF37',
                        'field': 'rb_subtraction',
                        'label': 'Maths: Subtraction Field',
                    },
                    {
                        'colour': '#D4AF37',
                        'field': 'rb_times',
                        'label': 'Maths: Times Field',
                    },
                    {
                        'colour': '#D4AF37',
                        'field': 'rb_division',
                        'label': 'Maths: Division Field',
                    },
                    {
                        'colour': '#D4AF37',
                        'field': 'rb_percentage',
                        'label': 'Maths: Percentage Field',
                    },
                ]
            )

    @staticmethod
    def _is_search_match(search_string, title):
        if search_string is None:
            return True
        return search_string.lower() in title.lower()



from django.urls import path

from advanced_report_builder.views.bar_charts import (
    BarChartBreakdownFieldModal,
    BarChartFieldModal,
    BarChartModal,
    BarChartShowBreakdownModal,
)
from advanced_report_builder.views.calendar import (
    CalendarDataSetDuplicateModal,
    CalendarDataSetModal,
    CalendarDescriptionDuplicateModal,
    CalendarDescriptionModal,
    CalendarModal,
)
from advanced_report_builder.views.custom import CustomModal
from advanced_report_builder.views.dashboard import (
    DashboardAddReportModal,
    DashboardModal,
    DashboardReportModal,
)
from advanced_report_builder.views.datatables.modal import (
    TableFieldModal,
    TableModal,
    TablePivotModal,
)
from advanced_report_builder.views.funnel_charts import (
    FunnelChartFieldModal,
    FunnelChartModal,
)
from advanced_report_builder.views.kanban import (
    KanbanDescriptionDuplicateModal,
    KanbanDescriptionModal,
    KanbanLaneDuplicateModal,
    KanbanLaneModal,
    KanbanModal,
)
from advanced_report_builder.views.line_charts import (
    LineChartFieldModal,
    LineChartModal,
)
from advanced_report_builder.views.pie_charts import PieChartFieldModal, PieChartModal
from advanced_report_builder.views.query_modal.modal import QueryModal, QueryOrderModal
from advanced_report_builder.views.reports import DuplicateReportModal
from advanced_report_builder.views.single_values import (
    QueryNumeratorModal,
    SingleValueModal,
    SingleValueShowBreakdownModal,
    SingleValueTableFieldModal,
)
from advanced_report_builder.views.targets.views import TargetModal

app_name = 'advanced_report_builder'


urlpatterns = [
    path('query/modal/query/<str:slug>', QueryModal.as_view(), name='query_modal'),
    path(
        'query/modal/query/order/<str:slug>',
        QueryOrderModal.as_view(),
        name='query_order_modal',
    ),
    path('table/modal/<str:slug>/', TableModal.as_view(), name='table_modal'),
    path(
        'table/modal/field/<str:slug>/',
        TableFieldModal.as_view(),
        name='table_field_modal',
    ),
    path(
        'table/modal/pivot/<str:slug>/',
        TablePivotModal.as_view(),
        name='table_pivot_modal',
    ),
    path(
        'single-value/modal/<str:slug>/',
        SingleValueModal.as_view(),
        name='single_value_modal',
    ),
    path(
        'single-value/modal/field/<str:slug>/',
        SingleValueTableFieldModal.as_view(),
        name='single_value_field_modal',
    ),
    path(
        'single-value/modal/numerator-field/<str:slug>/',
        QueryNumeratorModal.as_view(),
        name='single_value_numerator_modal',
    ),
    path(
        'single-value/show-breakdown/<str:slug>/',
        SingleValueShowBreakdownModal.as_view(),
        name='single_value_show_breakdown_modal',
    ),
    path('bar-chart/modal/<str:slug>/', BarChartModal.as_view(), name='bar_chart_modal'),
    path(
        'bar-chart/modal/field/<str:slug>/',
        BarChartFieldModal.as_view(),
        name='bar_chart_field_modal',
    ),
    path(
        'bar-chart/modal/breakdown/field/<str:slug>/',
        BarChartBreakdownFieldModal.as_view(),
        name='bar_chart_breakdown_field_modal',
    ),
    path(
        'bar-chart/show-breakdown/<str:slug>/',
        BarChartShowBreakdownModal.as_view(),
        name='bar_chart_show_breakdown_modal',
    ),
    path(
        'line-chart/modal/<str:slug>/',
        LineChartModal.as_view(),
        name='line_chart_modal',
    ),
    path(
        'line-chart/modal/field/<str:slug>/',
        LineChartFieldModal.as_view(),
        name='line_chart_field_modal',
    ),
    path('pie-chart/modal/<str:slug>/', PieChartModal.as_view(), name='pie_chart_modal'),
    path(
        'pie-chart/modal/field/<str:slug>/',
        PieChartFieldModal.as_view(),
        name='pie_chart_field_modal',
    ),
    path(
        'funnel-chart/modal/<str:slug>/',
        FunnelChartModal.as_view(),
        name='funnel_chart_modal',
    ),
    path(
        'funnel-chart/modal/field/<str:slug>/',
        FunnelChartFieldModal.as_view(),
        name='funnel_chart_field_modal',
    ),
    path('kanban/modal/<str:slug>/', KanbanModal.as_view(), name='kanban_modal'),
    path(
        'kanban/modal/lane/<str:slug>/',
        KanbanLaneModal.as_view(),
        name='kanban_lane_modal',
    ),
    path(
        'kanban/modal/lane/duplicate/<str:slug>/',
        KanbanLaneDuplicateModal.as_view(),
        name='kanban_lane_duplicate_modal',
    ),
    path(
        'kanban/modal/description/<str:slug>/',
        KanbanDescriptionModal.as_view(),
        name='kanban_description_modal',
    ),
    path(
        'kanban/modal/description/duplicate/<str:slug>/',
        KanbanDescriptionDuplicateModal.as_view(),
        name='kanban_description_duplicate_modal',
    ),
    path('calendar/modal/<str:slug>/', CalendarModal.as_view(), name='calendar_modal'),
    path('calendar/modal/data-set/<str:slug>/', CalendarDataSetModal.as_view(), name='calendar_data_set_modal'),
    path(
        'calendar/modal/lane/duplicate/<str:slug>/',
        CalendarDataSetDuplicateModal.as_view(),
        name='calendar_data_set_duplicate_modal',
    ),
    path(
        'calendar/modal/description/<str:slug>/', CalendarDescriptionModal.as_view(), name='calendar_description_modal'
    ),
    path(
        'calendar/modal/description/duplicate/<str:slug>/',
        CalendarDescriptionDuplicateModal.as_view(),
        name='calendar_description_duplicate_modal',
    ),
    path('custom/modal/<str:slug>/', CustomModal.as_view(), name='custom_modal'),
    path(
        'dashboard/report/<str:slug>/',
        DashboardReportModal.as_view(),
        name='dashboard_report_modal',
    ),
    path('dashboard/<str:slug>/', DashboardModal.as_view(), name='dashboard_modal'),
    path(
        'dashboard/add/<str:slug>/',
        DashboardAddReportModal.as_view(),
        name='add_dashboard_report',
    ),
    path(
        'duplicate/<str:slug>/',
        DuplicateReportModal.as_view(),
        name='duplicate_report_modal',
    ),
    path('target/modal/<str:slug>/', TargetModal.as_view(), name='target_modal'),
]
